import QtQuick 1.1
import qb.components 1.0
import qb.base 1.0
import "boilerStatus.js" as BoilerStatusJS

/*
 * BoilerstatusApp.qml
 *
 * Toon application for boiler parameters extraction. Extracted data are passed
 * on to BoilerstatusTile.qml
 * 
 * It requires just a working happ_thermstat implementation, on a qt-based 
 * toon gui. Should work with all fw. versions from 3.0.29 to at least 3.5.4.
 *
 * Fw 3.6.3 apparently no longer emits boiler data through the happ_thermstat
 * printTableInfo handler. 
 * 
 * 20170117: marcelr, first draft
 */

App {
    id: boilerStatusApp
    
    // These are the URL's for the QML resources from which our widgets will be instantiated.
    // By making them a URL type property they will automatically be converted to full paths,
    // preventing problems when passing them around to code that comes from a different path.
    
    property url tileUrl : "BoilerstatusTile.qml"
    property url thumbnailIcon: "./drawables/boilerstatus.png"
    
    property int unixTimestamp: BoilerStatusJS.getUnixTime()
    property bool haveOpenTherm

    /* boiler status parameters */

    property real boilerSetpoint
    property real remoteOverrideSetpoint
    property real maxModLevel
    property real maxBoilerPower
    property real roomTempSetpoint
    property real relModLevel
    property real boilerPressure
    property real roomTemp
    property real boilerOutTemp
    property real dhwTemp
    property real outsideTemp
    property real boilerInTemp
    property real solarStorageTemp
    property real maxDHWSetpoint
    property real maxCHWaterSetpoint
        
   
    function init() {
	registry.registerWidget( "tile", tileUrl, this, null, 
				 { thumbLabel: qsTr("Ketel status"), 
				   thumbIcon: thumbnailIcon, 
				   thumbCategory: "general", 
				   thumbWeight: 30, 
				   baseTileWeight: 10, 
				   thumbIconVAlignment: "center" } );

	}



    function updateBoilerStatus() {
	
	var xmlhttp = new XMLHttpRequest();
	var infoJson = {};

	xmlhttp.onreadystatechange = function() {

	    if  ( xmlhttp.readyState == 4 ) {
		if  ( xmlhttp.status == 200  ) {
		    
		    // transform text to proper JSON format.
		    
		    var txt = xmlhttp.responseText;
		    txt = txt.replace( "info = ", "{'info' : " ); 
		    txt = txt.replace( "\n]", "\n]}" );
		    txt = txt.replace( /'/g, "\"" ); //" extra quote to keep emacs happy
		    
		    infoJson = JSON.parse( txt );
		    var len = extractAndParse( infoJson );
		    
		}
	    }
	}
	xmlhttp.open( "GET", "http://127.0.0.1/happ_thermstat?action=printTableInfo",
		      true );
	xmlhttp.send();
	
    }

    


    function extractAndParse( data ) {

	/* 
	 * extractAndParse extracts data from a json info object, per OT ID, 
	 * and parses raw data according to the OT protocol.
	 */

	var len = data.info.length;
	var ID_l;
	var tstamp = 0;
	var val; 
	var val0;
	var val1;
	var i;
	var tag;
	var unit;
	
	/* check if we have Opentherm control or not. */
	
	if ( parseInt( data.info[0].updated ) === 0 ) {

	    /* boiler returns no data: timestamps are all 0 => bang-bang control. */
	    
	    console.log( "bang-bang-controlled boiler: no data to be found here" );
	    haveOpenTherm = 0;

	    return null;
	    
	} else {
	    
	    /* OpenTherm control, get all data. */

	    haveOpenTherm = 1;
	    
	    for ( i = 0; i < len; i++ ){
		
		ID_l   = parseInt( data.info[i].dataId );
		tstamp = parseInt( data.info[i].updated );

		/* 
		 * workaround for boiler modules that emit non-zero timestamps 
		 * even when there is no OT boiler in sight. 
		 * Allow a 5 minute grace period.
		 */

		if (( unixTimestamp - tstamp ) > 300 )
 		    haveOpenTherm = 0;


		switch ( ID_l ) {
		    
		    /* two bytes filled with switches, 16 bits total: */
		    
		case 0: /* status byte: high byte: master, low byte: slave                  */
		case 2: /* high byte: master configuration, low byte: master member ID      */
		case 3: /* high byte: slave configuration, low byte: slave member ID        */
		case 5: /* high byte: app-specific fault flags, low byte: OEM fault code    */
		case 6: /* high byte: remote par enable flag, low byte: remote par r/w flag */
		    
		    val  = Number( "0x" + data.info[i].value0 ) * 256;
		    val += Number( "0x" + data.info[i].value1 );
		    console.log( "timestamp = " + tstamp + ", ID = " + ID_l +  
				 ", val = " + val + ", bits = " + 
				 dec2bin( val ) );
		    
		    break;
		    
		    /* 
		     * f8.8: fixed point, 
		     * value = (signed high byte << 8 + unsigned low byte)/256 
		     */
		    
		case 1:  /* boiler control setpoint [deg C]      */
		case 9:  /* remote override setpoint [deg C]     */
		case 14: /* maximum modulation level setting [%] */ 
		case 15: /* maximum boiler power [kW]            */
		case 16: /* room temperature setpoint [deg C]    */
		case 17: /* relative modulation level [0-100%]   */
		case 18: /* boiler water pressure [bar]          */
		case 24: /* room temperature [deg C]             */
		case 25: /* boiler water temperature [deg C]     */
		case 26: /* DHW temperature [deg C]              */
		case 27: /* Outside temperature [deg C]          */
		case 28: /* return water temperature [deg C]     */
		case 29: /* solar storage temperature [deg C]    */
		case 56: /* max DHW setpoint [deg C]             */
		case 57: /* max CH water setpoint [deg C]        */
		    
		    val  = Number( "0x" + data.info[i].value0 );
		    if (val > 127)
			val = val - 256;
		    val += Number( "0x" + data.info[i].value1 ) / 256;
		    
		    switch (ID_l){
		    case 1:  /* boiler control setpoint [deg C]      */
			tag = "Boiler control setpoint = ";
			unit = "°C";
			boilerSetpoint = val;

/* 
 * Quick and dirty way to test for OT support. ID 1 is supported by all boilers.
 * When it's equal to zero, there's a strong possibility that your boiler
 * actually is a bang-bang controlled one, while the boiler module keeps 
 * spitting out a bunch of zeroes for no apparent reason.
 *
 * Not entirely fail safe, should do for now.
 */
//			if ( val != 0.0 )
//			    haveOpenTherm = 1;

			break;
		    case 9:  /* Remote override  setpoint [deg C]    */
			tag = "Remote override setpt   = ";
			unit = "°C";
			remoteOverrideSetpoint = val;
			break;
		    case 14: /* maximum modulation level setting [%] */ 
			tag = "Max. modulation level   = ";
			unit = "%";
			maxModLevel = val;
			break;
		    case 15: /* maximum boiler power [kW]            */
			tag = "Max. boiler power       = ";
			unit = "kW";
			maxBoilerPower = val;
			break;
		    case 16: /* room temperature setpoint [deg C]    */
			tag = "Room temp. setpoint     = ";
			unit = "°C";
			roomTempSetpoint = val;
			break;
		    case 17: /* relative modulation level [0-100%]   */
			tag = "Rel. modulation level   = ";
			unit = "%";
			relModLevel = val;
			break;
		    case 18: /* boiler water pressure [bar]          */
			tag = " Boiler water pressure  = ";
			unit = "bar";
			boilerPressure = val;
			break;
		    case 24: /* room temperature [deg C]             */
			tag = "Room temperature        = ";
			unit = "°C";
			roomTemp = val;
			break;
		    case 25: /* boiler water temperature [deg C]     */
			tag = "Boiler output temp      = ";
			unit = "°C";
			boilerOutTemp = val;
			break;
		    case 26: /* DHW temperature [deg C]              */
			tag = "DHW temperature         = ";
			unit = "°C";
			dhwTemp = val;
			break;
		    case 27: /* Outside temperature [deg C]          */
			tag = "Outside temperature     = ";
			unit = "°C";
			outsideTemp = val;
			break;
		    case 28: /* return water temperature [deg C]     */
			tag = "Return water temp.      = ";
			unit = "°C";
			boilerInTemp = val;
			break;
		    case 29: /* solar storage temperature [deg C]    */
			tag = "Solar storage temp.     = ";
			unit = "°C";
			solarStorageTemp = val;
			break;
		    case 56: /* max DHW setpoint [deg C]             */
			tag = "Max. DHW temperature     = ";
			unit = "°C";
			maxDHWSetpoint = val;
			break;
		    case 57: /* max CH water setpoint [deg C]        */
			tag = "Max CH setpoint temp.    = ";
			unit = "°C";
			maxCHWaterSetpoint = val;
			break;
		    default:
			tag = "Unknown ID  = ";
			break;
		    }		    
		    
		    console.log( "timestamp = " + tstamp + ", ID = " + ID_l + 
				 ",  val = " + val.toFixed(1) + unit );
		    break;
		    
		    /* two signed characters */
		    
		case 48: /* high byte: DHW temp high bound, low: DHW temp low bound */
		    val0 = Number( "0x" + data.info[i].value0 );
		    val1 = Number( "0x" + data.info[i].value1 );
		    console.log( "timestamp = " + tstamp + ", ID = " + ID_l +  
				 ", val0 = " + val0 + ", val1 = " + val1);
		    
		    break;
		    
		    /* unsigned short */
		    
		case 113: /* number of unsuccessful burner starts     */
		case 114: /* number of times flame signal was too low */
		case 115: /* OEM diagnostic code                      */
		case 116: /* number of burner starts                  */
		case 117: /* number of CH pump starts                 */
		case 118: /* number of DHW pump/valve starts          */
		case 119: /* number of burner starts for DHW          */
		case 120: /* burner operation hours                   */
		case 121: /* CH pump hours                            */
		case 122: /* DHW pump/valve hours                     */
		case 123: /* DHW burner hours                         */
		    
		    val  = Number( "0x" + data.info[i].value0 ) * 256;
		    val += Number( "0x" + data.info[i].value1 );
		    console.log( "timestamp = " + tstamp + ", ID = " + ID_l +  
				 ", val = " + val );
		    break;
		    
		    /* two unsigned characters */
		case 35: /* boiler fan speed setpoint/ actual (RPM/60 ) */ 	
		case 127: /*slave number and type */
		    
		    val0 = Number( "0x" + data.info[i].value0 );
		    val1 = Number( "0x" + data.info[i].value1 );
		    console.log( "timestamp = " + tstamp + ", ID = " + ID_l +  
				 ", val0 = " + val0 + ", val1 = " + val1);
		    break;
		    
		    /* parameters with unknown format */
		    
		case 73: /* OEM-specific code for ventilation / heat recovery */
		default:
		    
		    val  = Number( "0x" + data.info[i].value0 ) * 256;
		    val += Number( "0x" + data.info[i].value1 );
		    console.log( "timestamp = " + tstamp + ", ID = " + ID_l +  
				 ", val = " + val );
		    break;
		} // end of switch (ID) 
	    } // end of for loop over IDs
	    
	    return len;
	    
	} // end of if ... else
    }
    
    

    function dec2bin( dec ) {

	/* transform int to a set of printed bits. */

	return ( dec >>> 0 ).toString( 2 );
    } 
    

    Timer {
	id: datetimeTimer
	interval: 60000  // update every minute 
	triggeredOnStart: true
	running: true
	repeat: true
	onTriggered: updateBoilerStatus()
    }
   
}
